package com.bergerkiller.bukkit.nolagg;

import java.util.LinkedList;
import java.util.logging.Level;

import com.bergerkiller.bukkit.common.AsyncTask;
import com.bergerkiller.bukkit.common.Task;

public class OperationScheduler extends AsyncTask {
	
	public OperationScheduler() {
		super("NoLagg operation scheduler thread");
	}
	
	private static LinkedList<ThreadOperation> added = new LinkedList<ThreadOperation>();
	private static LinkedList<ThreadOperation> operations = new LinkedList<ThreadOperation>();
	private static LinkedList<ThreadOperation> finished = new LinkedList<ThreadOperation>();
	private static Task ticktask;
	private static AsyncTask schedulertask;

	public static void start(ThreadOperation operation) {
		synchronized (added) {
			added.offer(operation);
		}
		operation.start();
	}
	
	private static void startAll() {
		synchronized (added) {
			ThreadOperation op;
			while ((op = added.poll()) != null) {
				try {
				    op.start();
				    synchronized (operations) {
				    	operations.offer(op);
				    }
				} catch (Throwable t) {
					NoLagg.plugin.log(Level.SEVERE, "An error occured while starting a scheduled operation:");
					t.printStackTrace();
				}
			}
		}
	}
	private static void endAll() {
		synchronized (finished) {
			ThreadOperation op;
			while ((op = finished.poll()) != null) {
				try {
				    op.end();
				} catch (Throwable t) {
					NoLagg.plugin.log(Level.SEVERE, "An error occured while finishing a scheduled operation:");
					t.printStackTrace();
				}
			}
		}
	}
	private static void run(ThreadOperation op) {
		op.run();
		synchronized (finished) {
			finished.offer(op);
		}
	}
	
	public static void init() {
		schedulertask = new OperationScheduler().start(true);
		ticktask = new Task(NoLagg.plugin) {
			public void run() {
				startAll();
				endAll();
			}
		}.start(1, 1);
	}
	public static void deinit() {
		Task.stop(ticktask);
		ticktask = null;
		if (schedulertask != null) {
			schedulertask.stop().waitFinished();
			schedulertask = null;
		}
		synchronized (operations) {
			if (!operations.isEmpty()) {
				NoLagg.plugin.log(Level.INFO, "Performing remaining scheduled operations (" + operations.size() + ") ...");
			}
			startAll();
			for (ThreadOperation op : operations) {
				run(op);
			}
			operations.clear();
			endAll();
		}
	}
	
	private ThreadOperation operation;
	public void run() {
		synchronized (operations) {
			operation = operations.poll();
		}
		if (operation == null) {
			sleep(300);
		} else {
			try {
				run(operation);
			} catch (Throwable t) {
				NoLagg.plugin.log(Level.SEVERE, "An error occured while processing a scheduled operation:");
				t.printStackTrace();
			}
		}
	}

}
